/* 
 *  LevelPatcher Copyright (C) 2007, Donald Graft
 *
 *  LevelPatcher is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  LevelPatcher is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <windows.h>
#include "resource.h"
#include <stdio.h> 
#include <io.h>
#include <fcntl.h>
#include <commctrl.h>

char input_file[2048];
char output_file[2048];
int profile = 4; // main
int level = 8; // main
int inplace = 0;

HWND hWnd;
HANDLE hThread;
DWORD threadId;
DWORD WINAPI process(LPVOID n);
unsigned long pOldEditProc;

static void HandleDrop(HDROP drop)
{
    int count;
    char buf[MAX_PATH];

    count = DragQueryFile(drop, 0xFFFFFFFF, buf, MAX_PATH);

    if (count > 1)
		MessageBox(hWnd, "Multiple files dropped - only one file will be processed", "Warning", MB_OK | MB_ICONWARNING);

    if (DragQueryFile(drop, 0, buf, MAX_PATH) == 0)
	{
		MessageBox(hWnd, "Unable to get dropped file information", "Error", MB_OK | MB_ICONSTOP);
		return;
    }

    SetWindowText(GetDlgItem(hWnd, IDC_INPUT), buf);
	strcpy(input_file, buf);
	strcpy(output_file, buf);
	strcat(output_file, ".patched.m2v");
    SetWindowText(GetDlgItem(hWnd, IDC_OUTPUT), output_file);
}

static LRESULT CALLBACK NewEditSubWindowProc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{
    switch (msg)
    {
	case WM_DROPFILES:
	    HandleDrop((HDROP)wp);
	    return TRUE;

	default:
	    return CallWindowProc((WNDPROC)pOldEditProc, hwnd, msg, wp, lp);
    }
}

BOOL CALLBACK DlgProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	HICON hicon;
	switch(Message)
	{
		case WM_INITDIALOG:
			hWnd = hwnd;
			hicon = LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ICON)); 
			SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_SMALL, (LPARAM) hicon); 
			SendMessage(hwnd, (UINT) WM_SETICON, (WPARAM) ICON_BIG, (LPARAM) hicon); 
			SendDlgItemMessage(hwnd, IDC_PMAIN, BM_SETCHECK, 1, 0);
			SendDlgItemMessage(hwnd, IDC_LMAIN, BM_SETCHECK, 1, 0);
			pOldEditProc = SetWindowLong(GetDlgItem(hwnd, IDC_INPUT), GWL_WNDPROC, (LONG)NewEditSubWindowProc);
			break;

		case WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDC_PSIMPLE:
					profile = 5;
					break;
				case IDC_PMAIN:
					profile = 4;
					break;
				case IDC_PSNR:
					profile = 3;
					break;
				case IDC_PSPATIAL:
					profile = 2;
					break;
				case IDC_PHIGH:
					profile = 1;
					break;

				case IDC_LLOW:
					level = 10;
					break;
				case IDC_LMAIN:
					level = 8;
					break;
				case IDC_LHIGH1440:
					level = 6;
					break;
				case IDC_LHIGH:
					level = 4;
					break;

				case IDC_INPLACE:
					inplace = IsDlgButtonChecked(hwnd, IDC_INPLACE);
					if (inplace)
					{
						EnableWindow(GetDlgItem(hwnd, IDC_BROWSE_DEST), FALSE);
						EnableWindow(GetDlgItem(hwnd, IDC_OUTPUT), FALSE);
					}
					else
					{
						EnableWindow(GetDlgItem(hwnd, IDC_BROWSE_DEST), TRUE);
						EnableWindow(GetDlgItem(hwnd, IDC_OUTPUT), TRUE);
					}
					break;

				case IDC_CONVERT:
					GetDlgItemText(hwnd, IDC_INPUT, input_file, 2048);
					GetDlgItemText(hwnd, IDC_OUTPUT, output_file, 2048);
					SetDlgItemText(hwnd, IDC_STATUS, "Processing, please wait...");
					hThread = CreateThread(NULL, 32000, process, 0, 0, &threadId);
					break;

				case IDC_BROWSE:
					{
					OPENFILENAME ofn;
					input_file[0] = 0;
					ofn.lStructSize = sizeof(OPENFILENAME);
					ofn.hwndOwner = hwnd;
					ofn.hInstance = NULL;
					ofn.lpTemplateName = NULL;
					ofn.lpstrFilter = "All Files\0*.*\0\0";
					ofn.lpstrCustomFilter = NULL;
					ofn.nMaxCustFilter = 0;
					ofn.nFilterIndex = 1;
					ofn.lpstrFile = input_file;
					ofn.nMaxFile = 1024;
					ofn.lpstrFileTitle = NULL;
					ofn.nMaxFileTitle = 0;
					ofn.lpstrInitialDir = NULL;
					ofn.lpstrTitle = "Choose Input File";
					ofn.Flags = OFN_CREATEPROMPT;
					ofn.nFileOffset = 0;
					ofn.nFileExtension = 0;
					ofn.lpstrDefExt = NULL;
					ofn.lCustData = 0;
					ofn.lpfnHook = NULL;
					GetOpenFileName(&ofn);
					if (input_file[0] != 0)
					{
						SetWindowText(GetDlgItem(hwnd, IDC_INPUT), input_file);
						strcpy(output_file, input_file);
						strcat(output_file, ".patched.m2v");
						SetWindowText(GetDlgItem(hwnd, IDC_OUTPUT), output_file);
					}
					break;
					}

				case IDC_BROWSE_DEST:
					{
					OPENFILENAME ofn;
					output_file[0] = 0;
					ofn.lStructSize = sizeof(OPENFILENAME);
					ofn.hwndOwner = hwnd;
					ofn.hInstance = NULL;
					ofn.lpTemplateName = NULL;
					ofn.lpstrFilter = "All Files\0*.*\0\0";
					ofn.lpstrCustomFilter = NULL;
					ofn.nMaxCustFilter = 0;
					ofn.nFilterIndex = 1;
					ofn.lpstrFile = output_file;
					ofn.nMaxFile = 1024;
					ofn.lpstrFileTitle = NULL;
					ofn.nMaxFileTitle = 0;
					ofn.lpstrInitialDir = NULL;
					ofn.lpstrTitle = "Choose Output File";
					ofn.Flags = OFN_CREATEPROMPT;
					ofn.nFileOffset = 0;
					ofn.nFileExtension = 0;
					ofn.lpstrDefExt = NULL;
					ofn.lCustData = 0;
					ofn.lpfnHook = NULL;
					GetOpenFileName(&ofn);
					if (output_file[0] != 0)
					{
						SetWindowText(GetDlgItem(hwnd, IDC_OUTPUT), output_file);
					}
					break;
					}
			}
			break;

		case WM_CLOSE:
			SetWindowLong(GetDlgItem(hwnd, IDC_INPUT), GWL_WNDPROC, (LONG)pOldEditProc);
			EndDialog(hwnd, 0);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	int argc = __argc;
	char **argv = __argv;

	INITCOMMONCONTROLSEX InitCtrlEx;
	InitCtrlEx.dwSize = sizeof(INITCOMMONCONTROLSEX);
	InitCtrlEx.dwICC  = ICC_PROGRESS_CLASS;
	InitCommonControlsEx(&InitCtrlEx);
	return DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAIN), NULL, DlgProc);
}

// Defines for the start code detection state machine.
#define NEED_FIRST_0  0
#define NEED_SECOND_0 1
#define NEED_1        2

FILE *fp, *wfp;
int fd, wfd;

#define BUFFER_SIZE 32768
unsigned char buffer[BUFFER_SIZE];
unsigned char *Rdptr;
int Read;

void video_parser(void);

// For progress bar.
__int64 size;
__int64 data_count;

static int state, found;
int EOF_reached;

char stats[255];
char done[255];
unsigned int Start, Now, Elapsed;

void KillThread(void)
{
	Now = timeGetTime();
	if (Now >= Start)
		Elapsed = Now - Start;
	else
		Elapsed = (4294967295 - Start + Now) + 1;
	strcpy(done, "Done!  ");
	sprintf(stats,"[elapsed: %d sec]", Elapsed / 1000);
	strcat(done, stats);
	SetDlgItemText(hWnd, IDC_STATUS, done);
	ExitThread(0);
}

DWORD WINAPI process(LPVOID n)
{
	Start = timeGetTime();

	// Open the input file.
	if (inplace)
	{
		if (input_file[0] == 0 || (wfd = _open(input_file, _O_RANDOM | _O_BINARY | _O_RDWR)) == -1)
		{
			MessageBox(hWnd, "Could not open the input file!", "Error", MB_OK);
			KillThread();
		}
		// Read the first chunk of data.
		Read = _read(wfd, buffer, BUFFER_SIZE);
		Rdptr = buffer - 1;
	}
	else
	{
		if (input_file[0] == 0 || (fp = fopen(input_file, "rb")) == NULL)
		{
			MessageBox(hWnd, "Could not open the input file!", "Error", MB_OK);
			KillThread();
		}
		setvbuf(fp, NULL, _IOFBF, 0);
	}

	// Get the file size.
	fd = _open(input_file, _O_RDONLY | _O_BINARY | _O_SEQUENTIAL);
	size = _lseeki64(fd, 0, SEEK_END);
	_close(fd);

	// Open the output file.
	if (!inplace)
	{
		wfp = fopen(output_file, "wb");
		if (wfp == NULL)
		{
			char buf[80];
			sprintf(buf, "Could not open the output file\n%s!", output_file);
			MessageBox(hWnd, buf, "Error", MB_OK);
			KillThread();
		}
		setvbuf(wfp, NULL, _IOFBF, 0);
	}

	// Start converting.
	video_parser();

	return 0;
}

// Write a value in the file at an offset relative
// to the last read character. This allows for
// read ahead (write behind) by using negative offsets.
// The offset is ignored when not doing in-place operation.
_inline void put_byte(int offset, unsigned char val)
{
	__int64 save, backup;

	if (inplace)
	{
		// Save the file position.
		save = _lseeki64(wfd, 0, SEEK_CUR);
		// Calculate the amount to seek back to write this value.
		// offset is relative to the position of the last read value.
		backup = buffer + Read - Rdptr - offset;
		// Seek back.
		_lseeki64(wfd, -backup, SEEK_CUR);
		// Write the value.
		_write(wfd, &val, 1);
		// Restore the file position to be ready for the next
		// read buffer refill.
		_lseeki64(wfd, save, SEEK_SET);
	}
	else
	{
		fwrite(&val, 1, 1, wfp);
	}
}

// Get a byte from the stream. We do our own buffering
// for in-place operation, otherwise we rely on the
// operating system.
_inline unsigned char get_byte(void)
{
	unsigned char val;

	if (inplace)
	{
		if (Rdptr > &buffer[Read-2])
		{
			// Incrementing to the next byte will take us outside the buffer,
			// so we need to refill it.
			Read = _read(wfd, buffer, BUFFER_SIZE);
			Rdptr = buffer;
			if (Read < 1)
			{
				// No more data. Finish up.
				_close(wfd);
				SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 100, 0);
				KillThread();
			}
			// We return the first byte of the buffer now.
			val = *Rdptr;
		}
		else
		{
			// Increment the buffer pointer and return the pointed to byte.
			val = *++Rdptr;
		}
	}
	else
	{
		if (fread(&val, 1, 1, fp) != 1)
		{
			fclose(fp);
			if (wfp && !inplace) fclose(wfp);
			SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, 100, 0);
			KillThread();
		}
	}
	if (!(data_count++ % 524288) && (size > 0))
	{
		Now = timeGetTime();
		if (Now >= Start)
			Elapsed = Now - Start;
		else
			Elapsed = (4294967295 - Start) + Now + 1;
		SendDlgItemMessage(hWnd, IDC_PROGRESS, PBM_SETPOS, (UINT) (data_count * 100 / size), 0);
	}

	return val;
}

void video_parser(void)
{
	unsigned char val;
	
	// Inits.
	state = NEED_FIRST_0;
	found = 0;
	EOF_reached = 0;
	data_count = 0;

	// Let's go!
	while(1)
	{
		// Parse for start codes.
		val = get_byte();

	    if (!inplace) put_byte(0, val);

		switch (state)
		{
		case NEED_FIRST_0:
			if (val == 0)
				state = NEED_SECOND_0;
			break;
		case NEED_SECOND_0:
			if (val == 0)
				state = NEED_1;
			else
				state = NEED_FIRST_0;
			break;
		case NEED_1:
			if (val == 1)
			{
				found = 1;
				state = NEED_FIRST_0;
			}
			else if (val != 0)
				state = NEED_FIRST_0;
			break;
		}
		if (found == 1)
		{
			// Found a start code.
			found = 0;
			val = get_byte();
			if (!inplace) put_byte(0, val);

			if (val == 0xBA)
			{
				MessageBox(hWnd, "The input file must be an elementary\nstream, not a program stream.", "Error", MB_OK);
				KillThread();
			}
			else if (val == 0xB5)
			{
				val = get_byte();
				if ((val & 0xf0) == 0x10)
				{
					// Sequence extension
					val &= 0xf0;
					val |= profile;
					put_byte(0, val);
					val = get_byte();
					val &= 0x0f;
					val |= (level << 4);
					put_byte(0, val);
				}
				else
				{
					if (!inplace) put_byte(0, val);					
				}
			}
		}
	}
	return;
}
